
/**
 * Trida Uzel implementuje uzel binarniho vyhledavaciho stromu problemu kanibalu a misionaru
 * 
 * @author Martin ehnek
 * @version 1.0 4.10.2006
 */
public class Uzel
{
    private Uzel[] synove = new Uzel[5];
    private Uzel predek;
    private byte misionari;
    private byte kanibalove;
    private boolean lod;  //true pokud je lod na prvnim brehu
    private static boolean[][][] pouziteStavy = new boolean[4][4][2];   //pole uz pouzitych stavu 
    
    public byte getMisionari(){return misionari;}
    public byte getKanibalove(){return kanibalove;}
    public boolean getLod(){return lod;}
    public Uzel getPredek() {return predek;}
    public Uzel getSyn(byte i) {return synove[i];}
    private int abs(int number) {return  (number > 0) ? number : (-number) ;} 
    
   
    /**
     * @param  kanibalove   pocet kanibalu
     * @param  misionari   pocet misionaru
     * @param  lod   Pravda, pokud je lod na <B>prvnim</B> brehu
     */
    
    public Uzel(byte misionari, byte kanibalove, boolean lod, Uzel predek){
        this.misionari = misionari;
        this.kanibalove = kanibalove;
        this.lod = lod;
        this.predek = predek;
       }
    
    
    /**
     * Returns String reprezentujici stav
     * @return (pocet kanibalu, pocet misionaru, 1 pokud je lod na <B>prvnim</B> brehu, jinak 0)
     */
    public String toString()
      {
         return  "(" + misionari + "," + kanibalove + "," + (lod?1:0) +  ")" ; 
      }
    
    
    /**
     * Vrati pocet Synu daneho uzlu
     * @return Pocet synu
     */
    public byte pocetSynu()
     {
        byte i = 0;
        for (byte j = 0; j<5; j++)
          {if (this.getSyn(i) != null) 
             i++;
          }
      return i;
      }
        
   
    /**
     * Vygeneruje vsechny mozne potomky (objekty typu Uzel) a ulozi jejich reference do pole synove[]
     * 
     * @return Pocet opravdu vytvorenych potomku (max. 5)
     */
    public void generujSyny(){
       byte aktualBreh = (lod ? (byte)0 : (byte)3); // cislo, od ktereho se budou odecitat mis./kan., abych mohl pracovat stejne s obema brehy
       byte opacnyBreh = (lod ? (byte)3 : (byte)0);
       byte pM = (byte)(lod ? 1 : (-1)); //plusminus. Nasobi se tim konstanty, ktere pricitam(odecitam), abych mohl pocitat s obema brehy
       
       
       //1 misionar do lodky a na druhy breh  
         if ( ( abs(aktualBreh - misionari) > 0)  &&  //na aktualnim brehu aspon jeden misionar
              ( ( (abs(aktualBreh - misionari)- 1) >= abs(aktualBreh - kanibalove) ) || ((abs(aktualBreh - misionari)- 1) == 0) ) && // na brehu nebude ani po presunu prevaha kanibalu (jen pokud tam neni zadny misionar)
              ( ( (abs(opacnyBreh - misionari)+ 1) >= abs(opacnyBreh - kanibalove) ) || ((abs(opacnyBreh - misionari)+ 1) == 0) ) &&//podminka za poslednim || nikdy nebude true 
              !(pouziteStavy[misionari-1*pM][kanibalove][lod?0:1])            // Stav jaky by byl po presunu, tj. i opacny breh (1 - prvni breh(true), 0 - druhy(false) )
            )
            {
               synove[0] = new Uzel((byte)(misionari - 1*pM), kanibalove, !lod, this); 
               pouziteStavy[misionari-1*pM][kanibalove][lod?0:1] = true;
            }
        else synove[0] = null;
       
       
       //2 misionari do lodky a na druhy breh
       if ( ( abs(aktualBreh - misionari) > 1) &&  //na aktualnim brehu aspon dva misionari
            ( ( abs(aktualBreh - misionari) - 2 >= abs(aktualBreh - kanibalove) ) || ( abs(aktualBreh - misionari) - 2 == 0) ) &&// na aktualnim brehu nebude ani po presunu prevaha kanibalu
            ( ( abs(opacnyBreh - misionari) + 2 >= abs(opacnyBreh - kanibalove) ) || ( abs(opacnyBreh - misionari) + 2 == 0) ) &&// na opacnem brehu nebude ani po presunu prevaha kanibalu
            !(pouziteStavy[misionari - 2*pM][kanibalove][lod?0:1])
          )
          {
             synove[1] = new Uzel((byte)(misionari - 2*pM), kanibalove, !lod, this); 
             pouziteStavy[misionari - 2*pM][kanibalove][lod?0:1] = true;
          }
        else synove[1] = null;
       
       
       //1 kanibal do lodky a na druhy breh
       if ( ( abs(aktualBreh - kanibalove) > 0) &&  //na aktualnim brehu jeden kanibal
            ( ( abs(aktualBreh - misionari)  >= abs(aktualBreh - kanibalove) - 1 ) || (abs(aktualBreh - misionari) == 0) )&& // na brehu nebude ani po presunu prevaha kanibalu
            ( ( abs(opacnyBreh - misionari)  >= abs(opacnyBreh - kanibalove) + 1 ) || (abs(opacnyBreh - misionari) == 0) )&&// na brehu nebude ani po presunu prevaha kanibalu
            !(pouziteStavy[misionari][kanibalove-1*pM][lod?0:1])
          )
         {
             synove[2] = new Uzel(misionari, (byte)(kanibalove - 1*pM), !lod, this); 
             pouziteStavy[misionari][kanibalove-1*pM][lod?0:1] = true;
          }
        else synove[2] = null;
       
       
        //2 kanibalove do lodky a na druhy breh
       if ( ( abs(aktualBreh - kanibalove) > 1) &&  //na aktualnim brehu aspon dva kanibalove
            ( ( abs(aktualBreh - kanibalove) - 2 <= abs(aktualBreh - misionari) ) || (abs(aktualBreh - misionari) == 0) )&&// na brehu nebude ani po presunu prevaha kanibalu
            ( ( abs(opacnyBreh - kanibalove) + 2 <= abs(opacnyBreh - misionari) ) || (abs(opacnyBreh - misionari) == 0) )&& // na brehu nebude ani po presunu prevaha kanibalu
            !(pouziteStavy[misionari][kanibalove-2*pM][lod?0:1])
          )
          {
              synove[3] = new Uzel(misionari, (byte)(kanibalove - 2*pM), !lod, this); 
              pouziteStavy[misionari][kanibalove-2*pM][lod?0:1] = true; 
          }
        else synove[3] = null;
        
        //1 kanibal a 1 misionar
        if ( ( ( abs(aktualBreh - kanibalove) > 0) && (abs(aktualBreh - misionari) > 0) ) &&  //na aktualnim brehu aspon jeden kanibal a jeden misionar
             ( ( abs(aktualBreh - kanibalove) - 1 <= abs(aktualBreh - misionari) - 1 ) || (abs(aktualBreh - misionari) == 0) )&&// na brehu nebude ani po presunu prevaha kanibalu
             ( ( abs(opacnyBreh - kanibalove) + 1 <= abs(opacnyBreh - misionari) + 1 ) || (abs(opacnyBreh - misionari) == 0) )&&// na brehu nebude ani po presunu prevaha kanibalu
             !(pouziteStavy[misionari-1*pM][kanibalove-1*pM][lod?0:1]) 
           )
          {
              synove[4] = new Uzel((byte)(misionari - 1*pM), (byte)(kanibalove - 1*pM), !lod, this); 
              pouziteStavy[misionari-1*pM][kanibalove-1*pM][lod?0:1] = true;
          }
        else synove[4] = null;
                                                     
       }
       
}
